/* xkeyboard.c 
 *
 * Copyright (C) 1999-2000 Transmeta Corporation
 *
 * written by Nathan Laredo <nlaredo@transmeta.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/cursorfont.h>
#include <X11/Xmd.h>
#include <X11/Xmu/WinUtil.h>
#include <X11/extensions/XTest.h>
#include <X11/extensions/xtestext1.h>
#include "xkeyboard.h"
#ifdef NO_GRAPHICS
#include "nullkeycaps.h"
#else
#include "xkeycaps.h"
#endif

char *scanname[8][256];
int modifiermap[8][16];
int modindex[8];

Display *mydisplay;
Window mywindow, mykey[MAX_KEYS], ctrlwin, focuswin, focusparent,
	closewindow;
Window myborder[NUM_BORDERS], mylogo;
XWindowAttributes focusattribs, attribs;
int keyboard_active = 0;
GC mygc, mygcled, myact;
XGCValues gcvals;
XEvent myevent;
int myscreen;
Cursor mycursor, closecursor;
XSizeHints myhint;
XFontStruct *font_struct;
unsigned long myforeground, mybackground;
XSetWindowAttributes myattribs;	/* for window creation */
long myevents;
int keystate;
char tmpbuf[16];
int zoom = 0, ox, oy, sn, sd;
int screenwidth, screenheight;
Pixmap mykc[NUM_KEYCAPS];
enum { NORMAL, CLOSING } closingstate;

static int handle_error(Display *d, XErrorEvent *event)
{
	char errortext[96];
	XGetErrorText(mydisplay, event->error_code, errortext, 96);

	/* error is expected if focus window is destroyed */
	fprintf(stderr, "xkeyboard: %s (%d/%d)\n", errortext,
		event->request_code, event->minor_code);
	focuswin = focusparent = DefaultRootWindow(mydisplay);
	return 0;
}

void init_display(int argc, char **argv)
{
	XColor mycolor;

	if (!getenv("DISPLAY")) {
		fprintf(stderr, "DISPLAY not set.\n");
		exit(1);
	}
	if (!(mydisplay = XOpenDisplay(getenv("DISPLAY")))) {
		fprintf(stderr, "Unable to open display.\n");
		exit(1);
	}
	XSetErrorHandler(handle_error);
	XFlush(mydisplay);
#if 0
	/* for acting as a window manager */
	XSelectInput(mydisplay, DefaultRootWindow(mydisplay),
		SubstructureRedirectMask);
#endif
	myscreen = DefaultScreen(mydisplay);
	screenwidth = DisplayWidth(mydisplay, myscreen);
	screenheight = DisplayHeight(mydisplay, myscreen);
	focusparent = focuswin = DefaultRootWindow(mydisplay);
	mybackground = BlackPixel(mydisplay, myscreen);
	myforeground = WhitePixel(mydisplay, myscreen);
	myhint.width = 800;
	myhint.height = 240;
	myhint.max_width = 800;
	myhint.max_height = 240;
	myhint.min_width = 800;
	myhint.min_width = 240;
	myhint.x = 16-800;
	myhint.y = screenheight-240;
	myhint.flags = PSize | PMinSize | PMaxSize | PPosition;
	myattribs.event_mask = ButtonPressMask | VisibilityChangeMask |
		EnterWindowMask;
	myattribs.background_pixel = mybackground;
	myattribs.override_redirect = True;
	mywindow = XCreateWindow(mydisplay, DefaultRootWindow(mydisplay),
		myhint.x, myhint.y, myhint.width, myhint.height, 0,
		attribs.depth, InputOutput, attribs.visual,
		CWEventMask | CWBackPixel | CWOverrideRedirect, &myattribs);
	strcpy(tmpbuf, "xkeyboard");
	myattribs.event_mask = ButtonPressMask | VisibilityChangeMask |
		EnterWindowMask;
	XSetStandardProperties(mydisplay, mywindow, tmpbuf, tmpbuf, None,
	       argv, argc, &myhint);
	myhint.x = myhint.y = 0;
	myhint.width = myhint.height = 32;
	myhint.min_width = myhint.min_height = 32;
	myhint.flags = PSize | PMinSize | PPosition;
	gcvals.foreground = mybackground;
	gcvals.background = myforeground;
	mygc = XCreateGC(mydisplay, mywindow, GCForeground | GCBackground,
		&gcvals);
	font_struct = XLoadQueryFont(mydisplay,
		"-*-helvetica-medium-r-*-*-*-100-*-*-*-*-*-*");
	XSetFont(mydisplay, mygc, font_struct->fid);
	mycolor.red = 65535;
	mycolor.green = 65535;
	mycolor.blue = 0;
	mycolor.flags = DoRed | DoGreen | DoBlue;
	XAllocColor(mydisplay, DefaultColormap(mydisplay, myscreen), &mycolor);
	gcvals.foreground = mycolor.pixel;
	mygcled = XCreateGC(mydisplay, mywindow, GCForeground |
		GCBackground, &gcvals);
	XSetFont(mydisplay, mygcled, font_struct->fid);
	gcvals.foreground = mybackground;
	XMapRaised(mydisplay, mywindow);
}

Pixmap scale_pic(struct pgmdata *p, int kw, int kh, int num)
{
	Pixmap px;
	int x, y, w, h, pel, bpp, sx, sy, o, r, g, b;
	XImage *myimage;

	w = kw * sn / sd;
	h = kh * sn / sd;
	px = XCreatePixmap(mydisplay, mywindow, w, h,
		DefaultDepth(mydisplay, myscreen));
	if(!(myimage = XGetImage(mydisplay, px,
	   0, 0, w, h, AllPlanes, ZPixmap))) {
		fprintf(stderr, "XGetImage failed\n");
		exit(1);
	}
	if (p == &glassdownpgm || p == &glassuppgm ||
	    p == &closewindow_downpgm || p == &closewindow_uppgm) {
		/* don't resize magnifying glass */
		sx = sy = 256;
	} else {
		sx = p->width * 256 / w;
		sy = p->height * 256 / h; 
	}
	bpp = myimage->bits_per_pixel / 8;

	for (y = 0; y < h; y++)
		for (x = 0; x < w; x++) {
			o = myimage->bytes_per_line * y + x * bpp;
			pel = p->data[((sy * y)>>8) * p->width +
				((sx * x)>>8)];
			if (getenv("COLORKEYS") &&
			    (!num || num == 6 || num == 9 || num == 11)) {
				r = pel * 9 / 8;
				g = pel * 9 / 8;
				b = pel * 11 / 8;
				if (b > 255)
					b = 255;
				if (r > 255)
					r = 255;
				if (g > 255)
					g = 255;
			} else {
				r = g = b = pel;
			}
			if (myimage->depth == 16) {
				pel = ((r<<8)&63488)|((g<<3)&2016)|
					((b>>3)&31);
				myimage->data[o] = pel;
				myimage->data[o + 1] = (pel >> 8);
			} else if (myimage->depth == 15) {
				pel = ((r<<7)&31744)|((g<<2)&992)|
					((b>>3)&31);
				myimage->data[o] = pel;
				myimage->data[o + 1] = (pel >> 8);
			} else if (myimage->depth >= 24 ) {	/* 24/32 bpp */
				myimage->data[o] = r;
				myimage->data[o + 1] = g;
				myimage->data[o + 2] = b;
			} else {	/* 8bpp */
				myimage->data[o] = pel;
			}
	}
	XPutImage(mydisplay, px, mygc, myimage, 0, 0, 0, 0, w, h);
	XFree(myimage);
	return px;
}

static int kx_size[NUM_KEYCAPS] = {	/* hack */
	48, 64, 40, 32, 64, 32, 32, 80,
	80, 200, 64, 64, 48, 16, 16, 40, 40, 140, 32
};
static int ky_size[NUM_KEYCAPS] = {	/* hack */
	32, 32, 32, 32, 32, 64, 32, 32,
	32, 32, 32, 32, 32, 16, 16, 40, 40, 32, 32
};

void create_pixmaps(void)
{
	static int initialized = 0;
	int i;

	if (initialized) {
		for (i = 0; i < NUM_KEYCAPS; i++) {
			XFreePixmap(mydisplay, mykc[i]);
		}
	}
	for (i = 0; i < NUM_KEYCAPS; i++) {
		mykc[i] = scale_pic(mykclist[i], kx_size[i], ky_size[i], i);
	}
	initialized = 1;
}

void create_keys(void)
{
	int i, j, w, h;

	create_pixmaps();
	for (i = 0; i < MAX_KEYS; i++) {
		mykey[i] = XCreateSimpleWindow(mydisplay, mywindow,
			(ox + xk[i].x) * sn / sd, (oy + xk[i].y) * sn / sd,
			xk[i].w * sn / sd, xk[i].h * sn / sd,
			BORDER, mybackground, mybackground);
		XSetWindowBackgroundPixmap(mydisplay, mykey[i], mykc[xk[i].p]);
		XMapRaised(mydisplay, mykey[i]);
		XClearArea(mydisplay, mykey[i], 0, 0, 0, 0, True);
		XSelectInput(mydisplay, mykey[i], ButtonPressMask |
			ButtonReleaseMask | ExposureMask);
	}
	for (i = 0; i < NUM_BORDERS; i++) {
		w = mb[i].w * sn / sd;
		h = mb[i].h * sn / sd;
		if (!w)
			w = 4;
		if (!h)
			h = 4;
		myborder[i] = XCreateSimpleWindow(mydisplay, mywindow,
			(ox + mb[i].x) * sn / sd, (oy + mb[i].y) * sn / sd,
			w, h, 0, mybackground, mybackground);
		XMapRaised(mydisplay, myborder[i]);
	}
	keystate = 0;
	for (j = 0; j < 8; j++) {
		for (i = 0; i < 256; i++)
			scanname[j][i] = NULL;
		modindex[j] = 0;
	}
	ctrlwin =  XCreateSimpleWindow(mydisplay, mywindow,
		0, 202, 40, 40, 1, mybackground, myforeground);
	mylogo = XCreateSimpleWindow(mydisplay, mywindow,
		(ox + 608) * sn / sd, oy * sn / sd,
		140 * sn / sd, 32 * sn / sd,
		BORDER, mybackground, mybackground);
	closewindow = XCreateSimpleWindow(mydisplay, mywindow,
		800 - closewindow_uppgm.width, 0,
		closewindow_uppgm.width, closewindow_uppgm.height, 0,
		mybackground, mybackground);
	XSetWindowBackgroundPixmap(mydisplay, closewindow, mykc[NUM_KEYCAPS-5]);
	XSetWindowBackgroundPixmap(mydisplay, ctrlwin, mykc[NUM_KEYCAPS - 3]);
	XSetWindowBackgroundPixmap(mydisplay, mylogo, mykc[NUM_KEYCAPS - 2]);
	XSetWindowBackgroundPixmap(mydisplay, mywindow, mykc[NUM_KEYCAPS - 1]);
	XClearWindow(mydisplay, closewindow);
	XClearWindow(mydisplay, ctrlwin);
	XClearWindow(mydisplay, mylogo);
	XClearWindow(mydisplay, mywindow);
	XSelectInput(mydisplay, ctrlwin, ButtonPressMask | ButtonReleaseMask);
	XSelectInput(mydisplay, closewindow, ButtonPressMask);
	XSelectInput(mydisplay, mylogo, ButtonPressMask);
	XMapRaised(mydisplay, closewindow);
	XMapRaised(mydisplay, ctrlwin);
	XMapRaised(mydisplay, mylogo);
}

int get_keyboard_mapping(void)
{
	int i, j;
	int min_keycode, max_keycode, keysyms_per_keycode, keycode_count;
	XModifierKeymap *map;
	KeySym *keymap;
	char *tmp, *s;

	for (j = 0; j < 8; j++)
		for (i = 0; i < 16; i++)
			modifiermap[j][i] = 0;
	map = XGetModifierMapping(mydisplay);
	keycode_count = map->max_keypermod;
	for (i = 0; i < 8; i++) {
		for (j = 0; j < keycode_count && j < 16; j++) {
			modifiermap[i][j] =
				map->modifiermap[i * keycode_count + j];
		}
	}

	for (j = 0; j < 8; j++)
		for (i = 0; i < 256; i++)
			if (scanname[j][i] != NULL) {
				free(scanname[j][i]);
				scanname[j][i] = NULL;
			}
	XDisplayKeycodes(mydisplay, &min_keycode, &max_keycode);
	keycode_count = (max_keycode - min_keycode) + 1;
	if (!(keymap = XGetKeyboardMapping(mydisplay, min_keycode,
		keycode_count, &keysyms_per_keycode))) {
		fprintf(stderr, "Unable to get keyboard mapping table\n");
		return 1;
	}
	myevent.type = KeyPress;
	myevent.xkey.display = mydisplay;
	myevent.xkey.window = mywindow;
	myevent.xkey.same_screen = True;
	for (j = 0; j < keycode_count; j++) {
		for (i = 0; i < keysyms_per_keycode; i++) {
			KeySym ks = keymap[j * keysyms_per_keycode + i];
			if (ks != NoSymbol) {
				if ((s = XKeysymToString(ks)) != NULL) {
					tmp = malloc(16);
					memcpy(tmp, s, 16);
					tmp[15] = 0;
					scanname[i][j + min_keycode] = tmp;
					myevent.xkey.keycode = j + min_keycode;
					myevent.xkey.state = i & 3;
					XLookupString((XKeyEvent *) &myevent,
						tmpbuf, 16, &ks, NULL);
					if (tmpbuf[0] >= ' ' &&
					    tmpbuf[0] != 0x7f)
						memcpy(tmp, tmpbuf, 16);
					if ((tmp = strchr(&tmp[1], '_'))!=NULL)
						*tmp = 0;
				}
			}
		}
	}
	XFree((char *) keymap);
	return 0;
}

void redraw_window(Window exposed)
{
	int i, j, mod = 0, s;
	GC keygc;
	char *name;

	for (i = 0; i < MAX_KEYS; i++) {
		if (exposed == mykey[i]) {
			break;
		}
	}
	if (i == MAX_KEYS) {
		return;
	}
	s = xk[i].scancode;
	for (j = 0; j < 8; j++) {
		if (modifiermap[j][0] == s)
			break;
		if (modifiermap[j][1] == s)
			break;
		if (modifiermap[j][2] == s)
			break;
		if (modifiermap[j][3] == s)
			break;
	}
	keygc = mygc;
	if (j < 8) {
		if ((keystate & (1<< j)) && modindex[j] == i)
			keygc = mygcled;
	}

	for (mod = 0; mod < 4; mod++) {
		name = scanname[mod][s];
		if (!mod && !name && s == 94) {
			XUnmapWindow(mydisplay, mykey[i]);
		} else {
			if (!name)
				continue;
			if (islower(*name))
				continue;
			if (!mod && s == 94)
				XMapWindow(mydisplay, mykey[i]);
			if (name[0] == 'K' && name[1] == 'P')
				name += 3;
			XDrawString(mydisplay, mykey[i], keygc,
				8 * sn / sd + 8 * (mod & 2),
				26 * sn / sd - (11 * (mod & 1)) * sn / sd,
				name, strlen(name));
		}
	}
}

void resize_keys(void)
{
	int i, w, h;

	create_pixmaps();
	for (i = 0; i < MAX_KEYS; i++) {
		XMoveResizeWindow(mydisplay, mykey[i],
			(ox + xk[i].x) * sn / sd, (oy + xk[i].y) * sn / sd,
			xk[i].w * sn / sd, xk[i].h * sn / sd);
		XSetWindowBackgroundPixmap(mydisplay, mykey[i], mykc[xk[i].p]);
		/* generate a new expose event for window */
		XClearArea(mydisplay, mykey[i], 0, 0, 0, 0, True);
	}
	for (i = 0; i < NUM_BORDERS; i++) {
		w = mb[i].w * sn / sd;
		h = mb[i].h * sn / sd;
		if (!w)
			w = 4;
		if (!h)
			h = 4;
		XMoveResizeWindow(mydisplay, myborder[i],
			(ox + mb[i].x) * sn / sd, (oy + mb[i].y) * sn / sd,
			w, h);
	}
	XMoveResizeWindow(mydisplay, mylogo,
		(ox + 608) * sn / sd, oy * sn / sd,
		140 * sn / sd, 32 * sn / sd);
	XSetWindowBackgroundPixmap(mydisplay, closewindow,
		mykc[NUM_KEYCAPS - 5]);
	XSetWindowBackgroundPixmap(mydisplay, ctrlwin, mykc[NUM_KEYCAPS - 3]);
	XSetWindowBackgroundPixmap(mydisplay, mylogo, mykc[NUM_KEYCAPS - 2]);
	XSetWindowBackgroundPixmap(mydisplay, mywindow, mykc[NUM_KEYCAPS - 1]);
	XClearWindow(mydisplay, closewindow);
	XClearWindow(mydisplay, mywindow);
	XClearWindow(mydisplay, ctrlwin);
	XClearWindow(mydisplay, mylogo);
}

void do_control(void)
{
	zoom++;
	zoom &= 3;

	XFreeFont(mydisplay, font_struct);
	if (zoom == 0) {
		ox = 32; oy = 16; sn = 1; sd = 1;
		font_struct = XLoadQueryFont(mydisplay,
			"-*-helvetica-medium-r-*-*-*-100-*-*-*-*-*-*");
		XSetFont(mydisplay, mygc, font_struct->fid);
		XSetFont(mydisplay, mygcled, font_struct->fid);
	} else if (zoom == 1) {
		ox = 0; oy = -48; sn = 3; sd = 2;
		font_struct = XLoadQueryFont(mydisplay,
			"-*-helvetica-medium-r-*-*-*-120-*-*-*-*-*-*");
		XSetFont(mydisplay, mygc, font_struct->fid);
		XSetFont(mydisplay, mygcled, font_struct->fid);
	} else if (zoom == 3) {
		ox = -48; oy = -64; sn = 2; sd = 1;
		font_struct = XLoadQueryFont(mydisplay,
			"-*-helvetica-medium-r-*-*-*-180-*-*-*-*-*-*");
		XSetFont(mydisplay, mygc, font_struct->fid);
		XSetFont(mydisplay, mygcled, font_struct->fid);
	} else if (zoom == 2) {
		ox = -12; oy = -64; sn = 7; sd = 4;
		font_struct = XLoadQueryFont(mydisplay,
			"-*-helvetica-medium-r-*-*-*-120-*-*-*-*-*-*");
		XSetFont(mydisplay, mygc, font_struct->fid);
		XSetFont(mydisplay, mygcled, font_struct->fid);
	}
	resize_keys();
}

void resize_focus(int newsize)
{
	XSizeHints hints;
	long mask;
	int height_inc = 1, delta;

	if (!newsize) {
		XResizeWindow(mydisplay, focuswin,
			focusattribs.width, focusattribs.height);
		if (focusparent != DefaultRootWindow(mydisplay))
			XResizeWindow(mydisplay, focusparent,
				attribs.width, attribs.height);
	} else if (attribs.y + attribs.height + 240 > screenheight) {
		XGetWMNormalHints(mydisplay, focuswin, &hints, &mask);
		if (mask & PResizeInc) {
			height_inc = hints.height_inc;
			if (height_inc <= 0 || height_inc > (screenheight>>1))
				height_inc = 1;
		}
		delta = attribs.y + attribs.height + 240 - screenheight;
		delta += height_inc;
		/* snap to proper increment */
		delta = (delta / height_inc) * height_inc;
		if (delta <= 0 || delta >= focusattribs.height)
			return;
		XResizeWindow(mydisplay, focuswin,
			focusattribs.width, focusattribs.height - delta);
		if (focusparent != DefaultRootWindow(mydisplay))
			XResizeWindow(mydisplay, focusparent,
				attribs.width, attribs.height - delta);
	}
	XRaiseWindow(mydisplay, mywindow);
}

void new_focus(void)
{
	Window focus, newparent, root, parent, *children;
	int nchild;

	if (!keyboard_active) {
		focusparent = focuswin = DefaultRootWindow(mydisplay);
		XRaiseWindow(mydisplay, mywindow);
		return;
	}
	focus = focuswin;
	if (focuswin != DefaultRootWindow(mydisplay)) {
		resize_focus(0);
		if (focusparent != DefaultRootWindow(mydisplay)) {
			/* if reparented, lower parent window */
			XLowerWindow(mydisplay, focusparent);
		} else {
			XLowerWindow(mydisplay, focuswin);
		}
		XRaiseWindow(mydisplay, mywindow);
		XFlush(mydisplay);
	}
	if (XQueryTree(mydisplay, DefaultRootWindow(mydisplay),
			&root, &parent, &children, &nchild)) {
		while ((focus == focuswin || focus==mywindow ||
		       focus == focusparent) && nchild > 0) {
			nchild--;
			newparent = children[nchild];
			focus = XmuClientWindow(mydisplay, newparent);
			XGetWindowAttributes(mydisplay, focus, &focusattribs);
			if (focusattribs.map_state != IsViewable ||
			    focusattribs.class != InputOutput ||
			    !(focusattribs.all_event_masks & KeyPressMask)) {
				/* don't set to a window not in view */
				/* don't set to an inputonly window */
				/* don't set to a window not using keyboard */
				XLowerWindow(mydisplay, focus);
				focus = focuswin;
			}
		}
		XFree(children);
	}
	if (focus == focuswin) {
		focus = focuswin;
		newparent = focusparent;
		XGetWindowAttributes(mydisplay, focus, &focusattribs);
	}
	if (XQueryTree(mydisplay, focus, &root, &parent, &children, &nchild)) {
#if 0
		int i;
		fprintf(stderr, "focus=0x%08x, parent=0x%08x, root=0x%08x\n",
			focus, parent, root);
		for (i = 0; i < nchild; i++) {
			fprintf(stderr, "\t0x%08x\n", children[i]);
		}
#endif
		XFree(children);
	}
	if (parent == root) {
		/* not reparented or else no window manager */
		focusparent = root;
		attribs = focusattribs;
	} else {
		focusparent = newparent;
		/* focus window was reparented */
		XGetWindowAttributes(mydisplay, focusparent, &attribs);
	}
	focuswin = focus;
	XSetInputFocus(mydisplay, focuswin, RevertToPointerRoot,
		CurrentTime);
	if (focusparent != DefaultRootWindow(mydisplay)) {
		/* if reparented, raise parent window */
		XRaiseWindow(mydisplay, focusparent);
	} else {
		XRaiseWindow(mydisplay, focus);
	}
	resize_focus(1);
}

void do_button_event(void)
{
	int i, j, mod, sc, lockmap = 0;
	char *name;
	static Window lastkeywin = 0;
	Window win = myevent.xbutton.window;

	if (lastkeywin && myevent.type == ButtonRelease) {
		XSetWindowBorderWidth(mydisplay, lastkeywin, BORDER);
		win = lastkeywin;
		lastkeywin = 0;
	}
	if (lastkeywin) {	/* sanity */
		XSetWindowBorderWidth(mydisplay, lastkeywin, BORDER);
		lastkeywin = 0;
	}
	for (i = 0; i < MAX_KEYS; i++) {
		if (win == mykey[i])
			break;
	}
	if (i == MAX_KEYS) {
		/* unexpected event window */
		return;
	}
	sc = xk[i].scancode;
	for (j = 0; j < 8; j++) {
		if (modifiermap[j][0] == sc)
			break;
		if (modifiermap[j][1] == sc)
			break;
		if (modifiermap[j][2] == sc)
			break;
		if (modifiermap[j][3] == sc)
			break;
	}
	mod = 0;
	if (j < 8) {
		if (myevent.type == ButtonRelease)
			return;
		mod = (1 << j);
		if (keystate & mod) {
			keystate &= ~mod;
			lockmap |= mod;
			myevent.type = ButtonRelease;
			i = modindex[j];
			sc = xk[i].scancode;
			name = scanname[0][sc];
			if (name)
				XDrawString(mydisplay, mykey[i], mygc,
					8 * sn / sd, 26 * sn / sd,
					name, strlen(name));
		} else {
			keystate |= mod;
			lockmap |= mod;
			modindex[j] = i;
			name = scanname[0][sc];
			if (name)
				XDrawString(mydisplay, mykey[i], mygcled,
					8 * sn / sd, 26 * sn / sd,
					name, strlen(name));
		}
	}
	if (myevent.type == ButtonPress) {
		if (!mod)
			lastkeywin = mykey[i];
		myevent.type = KeyPress;
		XSetWindowBorderWidth(mydisplay, mykey[i], 2);
	} else {
		myevent.type = KeyRelease;
		XSetWindowBorderWidth(mydisplay, mykey[i], BORDER);
	}
	if ((lockmap & 0x92) && myevent.type == KeyRelease)
		myevent.type = KeyPress;
	XTestFakeKeyEvent(mydisplay, sc, (myevent.type == KeyPress), 0);
	if (myevent.type == KeyPress && !(mod)) {
		/* automagically turn off some modifiers */
		if (keystate & 0x01) { 	/* shift */
			i = modindex[0];
			sc = xk[i].scancode;
			XSetWindowBorderWidth(mydisplay, mykey[i], BORDER);
			XTestFakeKeyEvent(mydisplay, sc, 0, 0);
			name = scanname[0][sc];
			if (name)
				XDrawString(mydisplay, mykey[i], mygc,
					8 * sn / sd, 26 * sn / sd,
					name, strlen(name));
		}
		if (keystate & 0x04) {	/* control */
			i = modindex[2];
			sc = xk[i].scancode;
			XSetWindowBorderWidth(mydisplay, mykey[i], BORDER);
			XTestFakeKeyEvent(mydisplay, sc, 0, 0);
			name = scanname[0][sc];
			if (name)
				XDrawString(mydisplay, mykey[i], mygc,
					8 * sn / sd, 26 * sn / sd,
					name, strlen(name));
		}
		if (keystate & 0x08) { 	/* alt */
			i = modindex[3];
			sc = xk[i].scancode;
			XSetWindowBorderWidth(mydisplay, mykey[i], BORDER);
			XTestFakeKeyEvent(mydisplay, sc, 0, 0);
			name = scanname[0][sc];
			if (name)
				XDrawString(mydisplay, mykey[i], mygc,
					8 * sn / sd, 26 * sn / sd,
					name, strlen(name));
		}
		keystate &= ~0x0d;
	}
	if (lockmap & 0x92) {	/* Need to generate keyup events */
		if (lockmap & 0x02) {	/* Lock */
			XTestFakeKeyEvent(mydisplay,
				xk[modindex[1]].scancode, 0, 0);
		}
		if (lockmap & 0x10) {	/* Mod1 */
			XTestFakeKeyEvent(mydisplay,
				xk[modindex[4]].scancode, 0, 0);
		}
		if (lockmap & 0x80) {	/* Mod5 */
			XTestFakeKeyEvent(mydisplay,
				xk[modindex[7]].scancode, 0, 0);
		}
	}
	XFlush(mydisplay);
	return;
}

void activate_keyboard(int state)
{
	int i;
	if (state) {
		for (i = 16-800; i < 0; i += 64) {
			XMoveWindow(mydisplay, mywindow, i,
				    screenheight - 240);
		}
		XMoveWindow(mydisplay, mywindow, 0, screenheight - 240);
		if (focuswin == DefaultRootWindow(mydisplay)) {
			/* give current top-most window the focus */
			new_focus();
		} else {
			resize_focus(1);
		}
	} else {
		for (i = -16; i > 16-800; i -= 64) {
			XMoveWindow(mydisplay, mywindow, i,
				    screenheight - 240);
		}
		XMoveWindow(mydisplay, mywindow, 16-800, screenheight - 240);
		if (focuswin != DefaultRootWindow(mydisplay)) {
			resize_focus(0);
		}
		focuswin = DefaultRootWindow(mydisplay);
		XSetInputFocus(mydisplay, PointerRoot, RevertToPointerRoot,
				CurrentTime);
	}
}

void activate_closewindow(void)
{
	int rc;
	int done;
	Window dummy, target;
	int nuther_dummy;

	/* Depress the button */
	XSetWindowBackgroundPixmap(mydisplay, closewindow,
		mykc[NUM_KEYCAPS - 6]);
	XClearWindow(mydisplay, closewindow);

	/* Change the cursor so we know we're in killing mode */
	rc = XGrabPointer(mydisplay, mywindow, True,
		ButtonPressMask | ButtonReleaseMask, GrabModeAsync,
		GrabModeAsync, None, closecursor, CurrentTime);

	/* When I hit a ButtonPress event, send a Close event to
	   that window */
	if (rc != GrabSuccess) {
	} else {
		done = 0;
		while (!done) {
			XNextEvent(mydisplay, &myevent);
			if (myevent.type == ButtonPress) {
				XQueryPointer(mydisplay,
					DefaultRootWindow(mydisplay),
					&dummy, &target, &nuther_dummy,
					&nuther_dummy, &nuther_dummy,
					&nuther_dummy, &nuther_dummy);
				XUngrabPointer(mydisplay, CurrentTime);
				//target = XmuClientWindow(mydisplay, target);
				if (target && target != mywindow &&
				    target != DefaultRootWindow(mydisplay)) {
					XDestroyWindow(mydisplay, target);
					XFlush(mydisplay);
				}
				done = 1;
			}
		}
	}

	/* Restore the button in its up-state */
	XSetWindowBackgroundPixmap(mydisplay, closewindow,
		mykc[NUM_KEYCAPS - 5]);
	XClearWindow(mydisplay, closewindow);

}

int main(int argc, char **argv)
{
	int done = 0;

	init_display(argc, argv);
	ox = 32; oy = 16; sn = 1; sd = 1;
	create_keys();
	closingstate = NORMAL;
	mycursor = XCreateFontCursor(mydisplay, XC_hand2);
	closecursor = XCreateFontCursor(mydisplay, XC_pirate);
	if (get_keyboard_mapping())
		done++;
	/* Main event loop */
	while (!done) {
		XNextEvent(mydisplay, &myevent);
		switch (myevent.type) {
			case EnterNotify:
				if (!keyboard_active || focuswin ==
				    DefaultRootWindow(mydisplay)) {
					focuswin =
						DefaultRootWindow(mydisplay);
					new_focus();
				} else {
					XSetInputFocus(mydisplay, focuswin,
						RevertToPointerRoot,
						CurrentTime);
				}
				break;
			case ButtonPress:
				if (myevent.xbutton.window == mylogo) {
					if (myevent.xbutton.x > 104) {
						new_focus();
						break;
					}
					myevent.xbutton.window = mywindow;
				}
				if (myevent.xbutton.window == mywindow) {
					keyboard_active = 1 - keyboard_active;
					activate_keyboard(keyboard_active);
					break;
				}
				if (myevent.xbutton.window == ctrlwin) {
					/* "press" magnifying glass */
					XSetWindowBackgroundPixmap(mydisplay,
						ctrlwin,
						mykc[NUM_KEYCAPS - 4]);
					XClearWindow(mydisplay, ctrlwin);
					break;
				}
				if (myevent.xbutton.window == closewindow) {
					activate_closewindow();
					break;
				}
			case ButtonRelease:
				if (myevent.xbutton.window == ctrlwin) {
					do_control();
					break;
				}
				do_button_event();
				break;
			case Expose:
				if (!myevent.xexpose.count)
					redraw_window(myevent.xexpose.window);
				break;
			case VisibilityNotify:
				XRaiseWindow(mydisplay, mywindow);
				break;
			case UnmapNotify:
				if (myevent.xunmap.window == focuswin) {
					/* event window disappeared */
					focuswin =
						DefaultRootWindow(mydisplay);
					new_focus();
				}
				break;
			case CirculateNotify:
			case ConfigureNotify:
			case CreateNotify:
			case MapNotify:
			case ReparentNotify:
				break;
			case DestroyNotify:
				fprintf(stderr, "WM Terminated\n");
				done++;
				break;
			case MappingNotify:
				XRefreshKeyboardMapping(
					(XMappingEvent *) &myevent);
				if (get_keyboard_mapping())
					done++;
				redraw_window(mywindow);
				break;
		}
	}
	XFreeGC(mydisplay, mygc);
	XDestroyWindow(mydisplay, mywindow);
	XCloseDisplay(mydisplay);
	exit(0);
}
